'''
Security details calculation
'''
from .ledger_exec import LedgerExecutor


class SecurityDetails:
    '''
    The idea is to calculate the security details: 
    - average price (this will come with --average-lot-prices)
    - yield in the last 12 months
        - get the distributions "^income and :symbol$" in the last 12 months
        - divide by the current value
    '''
    def __init__(self, logger, symbol, currency):
        super().__init__()

        self.logger = logger
        self.symbol = symbol
        # The currency to use for all values
        self.currency = currency
    
    def calculate(self):
        '''
        The main method, which calculates everything.
        '''
        result = {}
        # result['message'] = ''
        # ledger = LedgerExecutor(self.logger)

        # lots
        # ledger_cmd = f'b ^Assets and :{self.symbol}$ --lots --no-total --depth 2'
        # lots = ledger.run(ledger_cmd)
        # result['lots'] = lots

        # average price
        # result['avg_price'] += 'N/A'

        # yield in the last 12 months
        result['yield'] = self.get_yield()

        # Gain/Loss
        result['g/l'] = self.get_gainloss()

        # income (demo)
        # income = self.get_income()
        # result['income'] = income

        return result

    def get_gainloss(self):
        ''' Get gain/loss from ledger '''
        ledger = LedgerExecutor(self.logger)
        # --collapse = -n
        # -G = gain/loss
        ledger_cmd = f'b ^Assets and :{self.symbol}$ -G -n -X {self.currency}'
        output = ledger.run(ledger_cmd)
        number = self.get_number_from_collapse_result(output)

        result = f"{number} {self.currency}"

        return result

    def get_yield(self):
        '''
        Calculate the yield in the last 12 months.
        This, of course is affected by the recent purchases, which affect the current value!
        '''
        from decimal import Decimal

        # get the income in the last 12 months
        income_str = self.get_income_balance()
        income = Decimal(income_str)
        #self.logger.debug(f'{self.symbol} gives `{income_str}` as income string')

        # turn into a positive number
        income = abs(income)

        # get the current value
        value_str = self.get_value_balance()
        value = Decimal(value_str)

        if value == 0:
            the_yield = 0
        else:
            the_yield = income * 100 / value

        result = f'{the_yield:.2f}%'
        return result

    def get_income(self):
        from datetime import date, timedelta

        yield_start_date = date.today() - timedelta(weeks=52)
        yield_from = yield_start_date.strftime("%Y-%m-%d")
        
        ledger = LedgerExecutor(self.logger)
        # the accound ends with the symbol name
        ledger_cmd = f'b ^Income and :{self.symbol}$ -b {yield_from} --flat --no-total'
        rows = ledger.run(ledger_cmd)
        rows = ledger.split_lines(rows)
        return rows

    def get_income_balance(self):
        ''' Gets the balance of income for the security '''
        from datetime import date, timedelta
        
        ledger = LedgerExecutor(self.logger)

        yield_start_date = date.today() - timedelta(weeks=52)
        yield_from = yield_start_date.strftime("%Y-%m-%d")
        
        # the accound ends with the symbol name
        # todo: use --collapse instead
        ledger_cmd = f'b ^Income and :{self.symbol}$ -b {yield_from} --flat -X {self.currency}'
        output = ledger.run(ledger_cmd)
        output = ledger.split_lines(output)
        #self.logger.debug(f'income lines for {self.symbol}: {output}')

        total = self.get_total_from_ledger_output(output)
        return total

    def get_value_balance(self):
        ''' Gets the current value of the security holdings in the given currency '''
        ledger = LedgerExecutor(self.logger)
        cmd = f"b ^Assets and :{self.symbol}$ -X {self.currency}"
        output = ledger.run(cmd)
        output = ledger.split_lines(output)
        value = self.get_total_from_ledger_output(output)
        return value

    def get_total_from_ledger_output(self, output):
        ''' Extract the total value from ledger output '''
        from cashiersync.ledger_output_parser import LedgerOutputParser

        parser = LedgerOutputParser()
        total_line = parser.get_total_lines(output)[0]
        #self.logger.debug(total_line)
        total_numeric = self.extract_total(total_line)
        return total_numeric

    def extract_total(self, total_line):
        ''' Gets the numeric value of the total from the ledger total line '''
        #self.logger.debug(total_line)
        
        # Extract the numeric value of the income total.
        total_parts = total_line.split(' ')
        total_numeric = total_parts[0]
        #self.logger.debug(f'total: {total_numeric}')
        # Remove thousand-separator
        total_numeric = total_numeric.replace(',', '') 

        # result = Decimal(total_numeric)
        result = total_numeric
        return result

    def get_number_from_collapse_result(self, ledger_output: str):
        ''' Parses a 1-line ledger result, when --collapse is used '''
        if not ledger_output:
            return 0

        # cleanup
        ledger_output = ledger_output.strip()
        # -1,139 EUR  Assets
        parts = ledger_output.split(' ')
        assert len(parts) == 4

        number = parts[0]
        number = number.replace(',', '') 

        return number



from json import encoder
from flask import Flask, request
from flask_cors import CORS #, cross_origin
import json
from .ledger_exec import LedgerExecutor

app = Flask(__name__)
CORS(app)
#cors = CORS(app)
#app.config['CORS_HEADERS'] = 'Content-Type'

@app.route("/")
def hello():
    return "Hello World!"

@app.route("/accounts")
def accounts():
    params = "accounts"
    ledger = LedgerExecutor(app.logger)
    result = ledger.run(params)
    
    return result

@app.route("/balance")
def balance():
    params = "b --flat --no-total"
    ledger = LedgerExecutor(app.logger)
    result = ledger.run(params)
    
    return result

@app.route("/currentValues")
def currentValues():
    root = request.args.get('root')
    currency = request.args.get('currency')
    params = f"b ^{root} -X {currency} --flat --no-total"

    ledger = LedgerExecutor(app.logger)
    result = ledger.run(params)
    
    #return f"current values for {root} in {currency}: {result}"
    return result

@app.route("/lots")
def lots():
    from .lots_parser import LotParser

    symbol = request.args.get('symbol')

    parser = LotParser(app.logger)
    result = parser.get_lots(symbol)

    return json.dumps(result)

@app.route('/payees')
def payees():
    ''' Send Payees as a simple list '''
    params = f"payees"
    ledger = LedgerExecutor(app.logger)
    result = ledger.run(params)
    return result

@app.route('/repo/pull', methods=['POST'])
def repo_pull():
    '''
    Pull the changes in the repository.
    Expects { repoPath: ... } parameter in JSON.
    '''
    from cashiersync.executor import Executor

    json = request.get_json()
    repo_path = json['repoPath']

    # Execute git pull and return all the console results.
    try:
        executor = Executor(app.logger)
        output = executor.run("git pull", repo_path)
    except Exception as e:
        output = f'Error: {str(e)}'

    return output

@app.route('/repo/commit', methods=['POST'])
def repo_commit():
    '''
    Pull the changes in the repository.
    Expects { repoPath: ... } parameter in JSON.
    '''
    from cashiersync.executor import Executor

    json = request.get_json()
    repo_path = json['repoPath']
    message = json['commitMessage']

    # Execute git pull and return all the console results.
    try:
        executor = Executor(app.logger)

        # Add all the changes.
        command = f"git add ."
        output = executor.run(command, repo_path)

        # Commit
        command = f'git commit -m "{message}"'
        output += executor.run(command, repo_path)
    except Exception as e:
        output = f'Error: {str(e)}'

    return output

@app.route('/repo/push', methods=['POST'])
def repo_push():
    '''
    Push the changes in the repository.
    Expects { repoPath: ... } parameter in JSON.
    '''
    from cashiersync.executor import Executor

    json = request.get_json()
    repo_path = json['repoPath']

    # Execute git pull and return all the console results.
    try:
        executor = Executor(app.logger)
        output = executor.run("git push", repo_path)
    except Exception as e:
        output = f'Error: {str(e)}'

    return output

@app.route('/append', methods=['POST'])
def append():
    ''' Append the text to the given (journal) file '''
    from cashiersync.executor import Executor
    import os.path

    json = request.get_json()
    file_path = json['filePath']
    content = json['content']

    # confirm that the file exists
    if not os.path.isfile(file_path):
        raise FileNotFoundError()
    
    # save
    with open(file_path, 'a') as destination:
        destination.write(content)

    return "Content written"

@app.route('/repo/status')
def repo_status():
    from cashiersync.executor import Executor
    
    repo_path = request.args.get('repoPath')
    try:
        executor = Executor(app.logger)
        output = executor.run("git status", repo_path)
    except Exception as e:
        output = f'Error: {str(e)}'

    return output

@app.route('/search', methods=['POST'])
def search_tx():
    ''' Search Transactions - Register '''
    from cashiersync.ledger_output_parser import LedgerOutputParser
    from cashiersync.model import TransactionEncoder

    #query = request.args.get('query')
    request_json = request.get_json()
    query = request_json['query']
    app.logger.debug(query)

    freeText = query['freeText']
    dateFrom = query['dateFrom']
    dateTo = query['dateTo']

    params = f"r"
    if freeText:
        params += f" {freeText}"
    if dateFrom:
        params += f" -b {dateFrom}"
    if dateTo is not None:
        params += f" -e {dateTo}"

    ledger = LedgerExecutor(app.logger)
    result = ledger.run(params)

    lines = result.split('\n')
    parser = LedgerOutputParser()
    lines = parser.clean_up_register_output(lines)

    txs = parser.get_rows_from_register(lines)
    #return result
    #encoded = TransactionEncoder().encode(txs)
    encoded = json.dumps(txs, cls=TransactionEncoder)
    #return json.dumps(txs)
    return encoded

@app.route('/securitydetails')
def security_details():
    ''' Displays the security details (analysis) '''
    from .sec_details import SecurityDetails

    symbol = request.args.get('symbol')
    currency = request.args.get('currency')

    x = SecurityDetails(app.logger, symbol, currency)
    result = x.calculate()

    return json.dumps(result)

@app.route('/transactions')
def transactions():
    ''' Fetch the transactions in account for the giver period '''
    from .transactions import LedgerTransactions

    account = request.args.get('account')
    dateFrom = request.args.get('dateFrom')
    dateTo = request.args.get('dateTo')

    assert account is not None
    assert dateFrom is not None
    assert dateTo is not None

    tx = LedgerTransactions()
    txs = tx.get_transactions(account, dateFrom, dateTo)

    return json.dumps(txs)

@app.route('/xact', methods=['POST'])
def xact():
    query = request.get_json()
    params = 'xact '

    if 'date' in query:
        date_param = query['date']
        params += date_param + ' '
    if 'freeText' in query:
        free_text = query['freeText']
        params += free_text

    #params = f'xact {date_param} {free_text}'

    ledger = LedgerExecutor(app.logger)
    try:
        result = ledger.run(params)
    except Exception as error:
        result = str(error)

    return result

@app.route('/about')
def about():
    ''' display some diagnostics '''
    import os
    cwd = os.getcwd()
    return f"cwd: {cwd}"

## Operations

@app.route('/shutdown')
def shutdown():
    #app.do_teardown_appcontext()

    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
        raise RuntimeError('Not running with the Werkzeug Server')
    func()

###################################

def run_server():
    """ Available to be called from outside """
    # use_reloader=False port=23948
    app.run(host='0.0.0.0', threaded=True, use_reloader=False, debug=True)
    # host='127.0.0.1' host='0.0.0.0'
    # , debug=True
    # Prod setup: 
    # debug=False


##################################################################################
if __name__ == '__main__':
    # Use debug=True to enable template reloading while the app is running.
    # debug=True <= this is now controlled in config.py.
    run_server()


'''
The ledger executor.
Runs ledger-cli to fetch the data.
'''

class LedgerExecutor:
    def __init__(self, logger):
        super().__init__()
        self.logger = logger

    def run(self, parameters):
        ''' Execute ledger command '''
        import subprocess
        from cashiersync.config import Configuration

        cfg = Configuration()
        cwd=cfg.ledger_working_dir

        command = f"ledger {parameters}"
        result = subprocess.run(command, shell=True, encoding="utf-8", capture_output=True,
            cwd=cwd)

        # if self.logger:
        #     self.logger.debug(result)

        if result.returncode != 0:
            output = result.stderr
            raise Exception('Ledger error', output)
        else:
            output = result.stdout
        
        #output = self.split_lines(output)

        return output

    def split_lines(self, output):
        rows = output.strip().split('\n')

        for i, item in enumerate(rows):
            rows[i] = rows[i].strip()
        
        return rows



'''
    Ledger Transactions
'''
from .ledger_exec import LedgerExecutor
from typing import List


class Transaction:
    ''' Transaction model '''
    def __init__(self):
        super().__init__()
        self.date = None
        self.payee = None
        self.account = None
        self.amount = None
        self.total = None


class LedgerTransactions:
    def __init__(self):
        super().__init__()

    def get_transactions(self, account, dateFrom, dateTo) -> List[str]:
        ''' Fetch historical transactions and return an array '''
        from cashiersync.json_helper import obj_to_dict

        ledger = LedgerExecutor(None)
        params = f'r "{account}" -b {dateFrom} -e {dateTo}'
        output = ledger.run(params)
        lines = ledger.split_lines(output)

        # Parse
        txs = []
        for line in lines:
            tx = self.__get_tx_from_line(line)
            dic = obj_to_dict(tx)
            txs.append(dic)

        return txs

    def __get_tx_from_line(self, line: str) -> Transaction:
        ''' parse ledger output line and create a Transaction object '''
        from cashiersync.ledger_output_parser import LedgerOutputParser

        parser = LedgerOutputParser()
        
        parts = line.split('  ')
        parts = parser.remove_blank_values_from_splits(parts)
        parts = parser.trim_all(parts)
        # date payee   account   amount  total
        date_payee = parts[0].split(' ')

        tx = Transaction()
        tx.date = date_payee[0]
        tx.payee = date_payee[1]
        tx.account = parts[1]
        tx.amount = parts[2]
        tx.total = parts[3]

        return tx




'''
    Models used in the app.
'''
#from json import JSONEncoder
import json
from decimal import Decimal


class RegisterRow:
    '''
    Represents the register transaction. 
    Also used for IB report comparison.
    '''
    def __init__(self):
        self.date: str = None
        self.payee: str = None
        self.account: str = None
        self.amount: Decimal = None
        self.currency: str = None

    def __repr__(self):
        return { 'date':self.date, 'payee':self.payee,
            'amount':self.amount, 'currency':self.currency }

    def __str__(self):
        return (f"Transaction(date='{self.date}', payee='{self.payee}'," + 
            f", amount={self.amount}, currency='{self.currency}')")

class Distribution:
    '''
    Represents the register transaction. 
    Also used for IB report comparison.
    Unused, but remains for the field reference.
    '''
    def __init__(self):
        self.date: str = None
        self.symbol: str = None
        self.type: str = None
        self.amount: Decimal = None
        self.currency: str = None

    def __repr__(self):
        return { 'date':self.date, 'symbol':self.symbol, 'type':self.type,
            'amount':self.amount, 'currency':self.currency }

    def __str__(self):
        return (f"Transaction(date='{self.date}', symbol='{self.symbol}'," + 
            f"type='{self.type}', amount={self.amount}, currency='{self.currency}')")

    def toJson(self):
        return json.dumps(self, default=lambda o: o.__dict__)


class TransactionEncoder(json.JSONEncoder):
    def default(self, object):
        if isinstance(object, RegisterRow):
            return object.__dict__
        elif isinstance(object, Decimal):
            return str(object)
        else:
            # call base class implementation which takes care of
            # raising exceptions for unsupported types
            return json.JSONEncoder.default(self, object)



'''
'''
from typing import List

def obj_to_dict(obj):
    dic = {}
    dic.update(vars(obj))
    return dic

def object_list_to_dicts(objects: List) -> List[str]:
    ''' Convert a list of objects to a list of dictionaries '''
    result = []

    for obj in objects:
        dic = obj_to_dict(obj)
        result.append(dic)

    return result
